/* Copyright (C) 2005-2007 Theodore A. Phelps */
/* Copyright (C) 2005-2007 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.  */

#if !defined(lint)
static const char cvsid[] = "@(#)$Id: speedwatch.c,v 1.222 2007/02/17 01:42:34 phelps Exp $";
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_PATHS_H
#include <paths.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_NETPACKET_PACKET_H
#include <netpacket/packet.h>
#endif
#ifdef HAVE_NET_ETHERNET_H
#include <net/ethernet.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYSLOG_H
#define SYSLOG_NAMES 1
#include <syslog.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#include <linux/atm.h>
#include <linux/atmdev.h>
#include <linux/atmbr2684.h>
#include <linux/if_pppox.h>
#include <linux/ppp_defs.h>

/*
 * Although the speedtouch driver and firmware are reasonably stable,
 * the ATM link will occasionally drop out, killing the network
 * connection.  Restarting usually involves removing the driver module
 * and re-inserting it to reset the state.
 *
 * Speedwatch should ensure that the speedtouch modem is restarted
 * when the ATM link drops out.  It will load the speedtch module and,
 * after a judicious pause, start pppd using the nodetach option to
 * keep it in the foreground.  By default, if pppd dies with an exit
 * code of 15 ("the link was terminated because the peer is not
 * responding to echo requests") then the speedtch module will be
 * unloaded and, after a few seconds, re-loaded, restarting the loop.
 *
 * If pppd dies with any other signal, the speedtch module will be
 * removed and this program will exit with pppd's exit code.
 *
 * If speedwatch receives a SIGINT or SIGTERM signal, it will unload
 * the speedtch module and quit with a 5 exit code, which corresponds
 * to pppd's exit code 5 ("pppd terminated because it was sent a
 * SIGINT, SIGTERM or SIGHUP).
 *
 * If speedwatch receives a SIGHUP signal, it will kill pppd and
 * behave as if it had died with an exit code of 15.
 */

/* pppd exit codes from pppd.h. */
#define EXIT_OK 0
#define EXIT_FATAL_ERROR 1
#define EXIT_OPTION_ERROR 2
#define EXIT_NOT_ROOT 3
#define EXIT_NO_KERNEL_SUPPORT 4
#define EXIT_USER_REQUEST 5
#define EXIT_LOCK_FAILED 6
#define EXIT_OPEN_FAILED 7
#define EXIT_CONNECT_FAILED 8
#define EXIT_PTYCMD_FAILED 9
#define EXIT_NEGOTIATION_FAILED 10
#define EXIT_PEER_AUTH_FAILED 11
#define EXIT_IDLE_TIMEOUT 12
#define EXIT_CONNECT_TIME 13
#define EXIT_CALLBACK 14
#define EXIT_PEER_DEAD 15
#define EXIT_HANGUP 16
#define EXIT_LOOPBACK 17
#define EXIT_INIT_FAILED 18
#define EXIT_AUTH_TOPEER_FAILED 19
#define EXIT_TRAFFIC_LIMIT 20
#define EXIT_CNID_AUTH_FAILED 21


/* The exit statuses for which we will attempt to send a PADT
 * packet, as a bitmask. */
#define PADT_EXIT_CODES \
    ((1 << EXIT_PEER_AUTH_FAILED) | \
     (1 << EXIT_AUTH_TOPEER_FAILED) | \
     (1 << EXIT_HANGUP))

/* LCP constants from RFC 1661. */
#define LCP_CONFIG_ACK 2
#define LCP_TERM_RQST 5
#define LCP_ECHO_RQST 9

/* Uncommon system call. */
extern long delete_module(const char *name, unsigned int flags);

/* The size of the buffer used to read from proc files. */
#define BUFFER_SIZE 1024

/* Macros for setting timeouts. */
#define WAKESET_SEC(sec, msec) sec
#define WAKESET_MSEC(sec, msec) msec
#define WAKESET(delay, res)                          \
    do {                                             \
        struct timeval ds_now, ds_diff;              \
        xgettimeofday(&ds_now);                      \
        ds_diff.tv_sec = WAKESET_SEC delay;          \
        ds_diff.tv_usec = 1000 * WAKESET_MSEC delay; \
        timeradd(&ds_now, &ds_diff, (res));          \
    } while (0)

#define WAKESETNONE(res)                        \
    do {                                        \
        (res)->tv_sec = -1;                     \
        (res)->tv_usec = -1;                    \
    } while (0);

#define WAKESETNOW(res) \
    memset((res), 0, sizeof(*res))

#define ASSERT(assertion)                                               \
    do {                                                                \
        if (!(assertion)) {                                             \
            xlog(LOG_ERR, "error: assertion \"%s\" failed %s:%d",       \
                 #assertion, __FILE__, __LINE__);                       \
            abort();                                                    \
        }                                                               \
    } while (0)

/* The time to wait after the line is up before starting pppd. */
#define LINE_UP_SETTLE_PAUSE (2, 0)

/* The time to wait after unloading the modules before trying to load
 * them again. */
#define UNLOAD_SETTLE_PAUSE (15, 0)

/* How long to pause when polling the modem status. */
#define SHORT_PAUSE (0, 500)


/* The amount of time to wait before deciding that the modem simply
 * doesn't exist. */
#define MODEM_INIT_TIMEOUT (30, 0)

/* The amount of time to wait before deciding that the ATM negotiation
 * isn't going to happen. */
#define MODEM_LINEUP_TIMEOUT (300, 0)

/* The amount of time to wait before giving up on a PPPoE session
 * packet. */
#define PPPoE_SESS_TIMEOUT (60, 0)

/* The amount of time to wait after pppd has crashed before starting a
 * new one. */
#define PPPD_RESTART_TIMEOUT (10, 0)

/* When reading the modem status file, the number of times that we
 * fail to interpret it before we give up. */
#define MAX_MALFORMED 5

/* Maximum number of authentication failures without a successful
 * connection before we give up. */
#define MAX_AUTH_ERRORS 3

/* Arguments for modprobe. */
#define MODPROBE_ARGS "-s"

/* Cleanup flags. */
#define CU_MODULES (1 << 0)
#define CU_BRIDGE (1 << 1)

/* The command-line options. */
#define OPTIONS "a:e:Ff:hi:km:p:su:v:V"

/* Paths. */
#ifndef ROOTDIR
#error ROOTDIR is undefined
#define ROOTDIR ""
#endif

#define PATH_PIDFILE ROOTDIR "/var/run/speedwatch.pid"
#define PATH_PPPD ROOTDIR "/usr/sbin/pppd"
#define PATH_PROC ROOTDIR "/proc"
#define PATH_PROC_MODULES PATH_PROC "/modules"
#define PATH_PROC_MODPROBE PATH_PROC "/sys/kernel/modprobe"
#define PATH_PROC_ATM PATH_PROC "/net/atm"
#define PATH_PROC_DEVICES PATH_PROC_ATM "/devices"
#define PATH_IFCONFIG ROOTDIR "/sbin/ifconfig"

/* Macro to log state changes. */
#define SET_STATE(var, state) \
    do { \
        xlog(LOG_DEBUG, "changing to state " #state); \
        var = state; \
    } while (0)

/* Parser states used by parse_pvc. */
typedef enum  {
    ST_START,
    ST_NUM,
    ST_ZERO,
    ST_END
} pvc_state_t;

/* Connection types. */
typedef enum {
    CT_PPPoA,
    CT_PPPoE
} conntype_t;

/* The FSM states. */
typedef enum
{
    /* Initial state: module and firmware not loaded, pppd not running. */
    FSM_START,

    /* Module loaded.  Waiting for device status file to appear. */
    FSM_FIND_STATUS,

    /* Module loaded.  Waiting for ATM link to be established. */
    FSM_CARRIER,

    /* Start pppd. */
    FSM_PPPD,

    /* Detach. */
    FSM_SESSION,

    /* Wait for pppd to die. */
    FSM_UP,

    /* Waiting for a PPPoE session packet so we can send a PADT. */
    FSM_PADT,

    /* Waiting for the pppd child process to die before we exit. */
    FSM_REAP
} fsm_state_t;

/* The speedtouch modem status. */
typedef enum
{
    /* The status file can't be found.  Either the speedtch module
     * hasn't been loaded or no modem was detected. */
    STATUS_NO_MODEM,

    /* The modem is disconnected. */
    STATUS_DISCONNECTED,

    /* Modem found, firmware loaded, line is down. */
    STATUS_LINE_DOWN,

    /* Modem found, firmware loaded, line status unknown. */
    STATUS_LINE_UNKNOWN,

    /* Modem found, firmware loaded, line is up. */
    STATUS_LINE_UP,
} status_t;

/* Internal error codes and their textual format. */
char *int_err_strings[] =
{
    "No error",
#define EPARSE -1
    "Malformed input",
#define EBADEXIT -2
    "Exit code indicates failure"
#define MIN_ERROR -3
};

/* Map status names to strings. */
char *status_names[] =
{
    "modem not found",
    "modem is disconnected",
    "line is down",
    "line status unknown",
    "line is up"
};

/* An RFC 2684 bridge. */
struct br2684
{
    /* The socket which represents the bridge. */
    int sock;

    /* The bridge device's name. */
    char ifname[IFNAMSIZ + 1];

    /* The local MAC address. */
    unsigned char mac_local[IFHWADDRLEN];

    /* The peer's MAC address. */
    unsigned char mac_peer[IFHWADDRLEN];

    /* The session_id or 0 if none found yet. */
    uint16_t session_id;
};

/* A PADT packet. */
struct padt_pkt
{
    struct ethhdr eth;
    struct pppoe_hdr pppoe;
} __attribute__ ((packed));

/* A LCP header. */
struct lcphdr
{
    unsigned char code;
    unsigned char identifier;
    unsigned short length;
} __attribute__ ((packed));

/* An LCP packet. */
struct lcp_pkt
{
    struct ethhdr eth;
    struct pppoe_hdr pppoe;
    unsigned short proto;
    struct lcphdr lcp;
} __attribute__ ((packed));

/* The finite state machine state information. */
struct fsm
{
    /* The current FSM state. */
    fsm_state_t state;

    /* Non-zero if exiting on errors. */
    unsigned int exit_on_error:1;

    /* Non-zero if we intend to detach. */
    unsigned int daemonize:1;

    /* Non-zero if we should only detach in response to SIGUSR1. */
    unsigned int detach_on_signal:1;

    /* The list of modules to load. */
    const char *mod_list;

    /* The interface number of the ATM device to use, or -1 if the
     * choice of device constrained by its interface number. */
    int atm_itf;

    /* The MAC address of the ATM device to use, or NULL if the choice
     * of device isn't constrained by the MAC address. */
    const char *mac_addr;

    /* The number of args to pass to pppd. */
    int pppd_argc;

    /* The args to pass to pppd. */
    char **pppd_argv;

    /* The time at which we should give up on the current state, reset
     * something and try again. */
    struct timeval expire;

    /* The time at which we should wake up and try again. */
    struct timeval wake;

    /* The path to the usbatm modem status file. */
    char path_modem_status[128];

    /* The modem status when last sampled. */
    status_t status;

    /* The number of consecutive times we've failed to parse the
     * usbatm modem status file. */
    int malformed_count;

    /* The pppd process id or (pid_t)-1 if pppd is not running. */
    pid_t pppd_pid;

    /* The connection type (PPPoA or PPPoE). */
    conntype_t conntype;

    /* The ATM address for the PPPoE bridge. */
    struct sockaddr_atmpvc atm_addr;

    /* The PPPoE encapsulation method to use (BR2684_ENCAPS_LLC or
     * BR2684_ENCAPS_VC). */
    int encap;

    /* The RFC 2684 bridge. */
    struct br2684 bridge;

    /* The packet capture socket. */
    int pcap_sock;

    /* The code to exit with. */
    int exit_code;
};


/* The name of the executable, stripped of leading path. */
static char *progname;

/* The path to the modprobe executable. */
static char *path_modprobe;

/* How much noise to make */
static int verbosity = LOG_WARNING;

/* Non-zero if we should log to stderr rather than syslog. */
static int log_to_stderr;

/* Where to record our pid */
static char *pid_file = PATH_PIDFILE;

/* The id of the child process, used to forward signals. */
static pid_t child_pid;

/* The last signal received. */
static int last_signum;

/* Whitespace characters. */
static const char *const ws = " \t\n";

/* Signals which should be forwarded to the descandant process. */
const int forwarded_signals[] = {
    SIGHUP, SIGINT, SIGQUIT, SIGTERM, 0
};

/* Signals which should be blocked while traversing state machine
 * states. */
const int blocked_signals[] = {
    SIGHUP, SIGINT, SIGUSR1, SIGTERM, 0
};

static void
usage()
{
    fprintf(stderr,
            "usage: %s [OPTION...] pppd-arg...\n"
            "    -a [itf.]vpi.vci\tUse PPPoE with given parameters\n"
            "    -e llc|vcmux\tEnable virtual circuit encapsulation for PPPoE [llc]\n"
            "    -u mod-list\t\tComma-separated list of modules\n"
            "    -f facility\t\tWrite syslog messages to facility [daemon]\n"
            "    -i num\t\tUse the ATM device on interface num\n"
            "    -k\t\tKeep going in the face of recoverable errors\n"
            "    -m addr\t\tUse the ATM device with mac address addr\n"
            "    -p file\t\tRecord process id in file [" PATH_PIDFILE "]\n"
            "    -F\t\t\tRun in the foreground\n"
            "    -s\t\t\tDetach on SIGUSR1\n"
            "    -v num\t\tSet the verbosity level [warning]\n"
            "    -h\t\t\tPrint this help message\n"
            "    -V\t\t\tPrint version information and exit\n",
            progname);
}

/* Write a message to syslog or stderr, as appropriate.  Note that
 * xlog will append a linefeed if necessary, so the format string
 * should not end with one. */
static void
xlog(int priority, char *format, ...)
{
    char buffer[1024];
    va_list ap;

    /* Bail if the log level is too high. */
    if (verbosity < priority) {
        return;
    }

    va_start(ap, format);
    
    /* If we're running in the foreground then print to stderr.
     * Otherwise write to syslog. */
    if (log_to_stderr) {
        vsnprintf(buffer, sizeof(buffer), format, ap);
        fprintf(stderr, "%s: %s\n", progname, buffer);
    } else {
        vsyslog(priority, format, ap);
    }

    va_end(ap);
}

/* Convert an error code into a string. */
static const char *
xstrerror(int err)
{
    if (MIN_ERROR < err && err < 0) {
        return int_err_strings[-err];
    } else {
        return strerror(err);
    }
}

/* Get the current time, or die trying. */
static void
xgettimeofday(struct timeval *tv)
{
    if (gettimeofday(tv, NULL) < 0) {
        xlog(LOG_ERR, "error: gettimeofday failed: %s", xstrerror(errno));
        exit(1);
    }
}

/* Handler for blocked signals. */
static void
handle_signal(int signum)
{
    ASSERT(!last_signum);
    xlog(LOG_DEBUG, "received signal %d", signum);
    last_signum = signum;
}

/* Forward a signal to the descendant process. */
static void
forward_signal(int signum)
{
    xlog(LOG_DEBUG, "forwarding signal %d to pid=%ld",
         signum, (long)child_pid);
    if (kill(child_pid, signum) < 0) {
        xlog(LOG_ERR, "error: kill failed: %s\n", xstrerror(errno));
    }
}

/* Read a pid from a file descriptor. */
static int
read_pid(int fd)
{
    pid_t pid;
    ssize_t len;
    int i;

    i = 0;
    while (i < sizeof(pid)) {
        if ((len = read(fd, (char *)&pid + i, sizeof(pid) - i)) < 0) {
            return (pid_t)-1;
        }
        /* Unexpected EOF. */
        if (len == 0) {
            errno = ENODATA;
            return (pid_t)-1;
        }

        i += len;
    }

    return pid;
}

/* Invoke a command and wait for it to exit. */
static int
invoke(const char *path, char *const argv[])
{
    pid_t pid;
    int status;
    sigset_t sigset;

    /* There should be no pending signals. */
    ASSERT(last_signum == 0);

    /* Spawn a subprocess */
    if ((pid = fork()) == (pid_t)-1) {
        return errno;
    }

    if (pid == 0) {
        /* We're the child: exec the program. */
        execv(path, argv);
        xlog(LOG_ERR, "error: unable to exec %s", path);
        _exit(1);
    }

    /* Wait for the exit signal. */
    ASSERT(sigfillset(&sigset) == 0);
    ASSERT(sigdelset(&sigset, SIGCHLD) == 0);
    while (last_signum == 0) {
        ASSERT(sigsuspend(&sigset) < 0);
        if (errno != EINTR) {
            return errno;
        }
    }
    ASSERT(last_signum == SIGCHLD);
    last_signum = 0;

    /* We're the parent: wait for the child to die. */
    if (waitpid(pid, &status, 0) == (pid_t)-1) {
        return errno;
    }

    /* Check for a happy exit code */
    if (!WIFEXITED(status) || WEXITSTATUS(status) != 0) {
        /* The subprocess exited with a non-zero exit code or was
         * killed with a signal. */
        if (WIFEXITED(status)) {
            xlog(LOG_INFO, "error: %s exited with code %d",
                 path, WEXITSTATUS(status));
        } else {
            xlog(LOG_INFO, "error: %s kill by signal %d",
                 path, WTERMSIG(status));
        }

        return EBADEXIT;
    }

    return 0;
}

static int
invokev(const char *path, char *arg, ...)
{
    char **argv;
    int argc, i;
    va_list ap;

    /* Determine how many arguments were provided. */
    ASSERT(arg != NULL);
    argc = 2;
    va_start(ap, arg);
    while (va_arg(ap, const char *) != NULL) {
        argc++;
    }
    va_end(ap);

    /* Allocate room for the arg list. */
    if ((argv = alloca(argc * sizeof(arg))) == NULL) {
        return errno;
    }

    /* Generate argv. */
    i = 0;
    argv[i++] = (char *)arg;
    va_start(ap, arg);
    while ((argv[i++] = va_arg(ap, char *)) != NULL);
    va_end(ap);

    ASSERT(argc == i);
    return invoke(path, argv);
}

/* Check the modem line status.  Returns 0 to indicate that all is
 * well, -1 to say that it doesn't understand the status file and
 * errno to indicate any other error. */
static int
modem_get_status(char *path, status_t *status_out)
{
    FILE *file;
    char buffer[BUFFER_SIZE];
    char *point;
    char *word;
    int i, err;

    /* Open the stats file. */
    if ((file = fopen(path, "r")) == NULL) {
        if (errno == ENOENT) {
            *status_out = STATUS_NO_MODEM;
            return 0;
        } else {
            return errno;
        }
    }

    /* Read the first four lines, discarding all but the last. */
    for (i = 0; i < 4; i++) {
        if ((fgets(buffer, sizeof(buffer), file)) == NULL) {
            if (ferror(file)) {
                err = errno;
                fclose(file);
                return err;
            }

            /* Premature end of file. */
            fclose(file);
            return EPARSE;
        }
    }

    /* We're done with the file, so close it. */
    if (fclose(file) < 0) {
        return errno;
    }

    /* Extract the first word of the line. */
    point = buffer;
    word = strsep(&point, ws);

    /* Is the device disconnected? */
    if (strcmp(word, "Disconnected") == 0) {
        *status_out = STATUS_DISCONNECTED;
        return 0;
    }

    /* Otherwise the first word must be Line. */
    if (strcmp(word, "Line") != 0) {
        return EPARSE;
    }

    /* According to usbatm.c, the next word(s) should be
     * "up", "down" or "state unknown". */
    word = strsep(&point, ws);
    if (strcmp(word, "up") == 0) {
        *status_out = STATUS_LINE_UP;
        return 0;
    }

    if (strcmp(word, "down") == 0) {
        *status_out = STATUS_LINE_DOWN;
        return 0;
    }

    if (strcmp(word, "state") == 0) {
        /* Assume that the next word is "unknown". */
        *status_out = STATUS_LINE_UNKNOWN;
        return 0;
    }

    /* Can't parse source file. */
    return EPARSE;
}

/* Get the path to modprobe from /proc. */
static int
find_modprobe()
{
    int err;
    char buffer[BUFFER_SIZE];
    FILE *file;
    char *point;

    /* Open /proc/sys/kernel/modprobe. */
    if ((file = fopen(PATH_PROC_MODPROBE, "r")) == NULL) {
        return errno;
    }
    
    /* Read the one line in it. */
    if (fgets(buffer, sizeof(buffer), file) == NULL) {
        err = errno;
        fclose(file);
        return err;
    }

    /* Kill the linefeed. */
    if ((point = strchr(buffer, '\n')) != NULL) {
        *point = 0;
    }

    /* Duplicate the string. */
    if ((path_modprobe = strdup(buffer)) == NULL) {
        err = errno;
        fclose(file);
        return err;
    }

    /* Close it. */
    if (fclose(file) < 0) {
        return errno;
    }

    return 0;
}

/* Canonicalize a module name. */
static void
module_canonical_name(const char *name,
                      char *canon_out,
                      size_t canon_size)
{
    const char *in = name;
    char *out = canon_out;

    while (*in != 0 && out < canon_out + canon_size - 1) {
        if (*in == '-') {
            *out = '_';
        } else {
            *out = *in;
        }

        in++;
        out++;
    }

    *out = 0;
}

/* See if a module is loaded. */
static int
module_is_loaded(const char *name, int *found_out)
{
    FILE *file;
    char buffer[BUFFER_SIZE];
    char *module_name;
    char *point;
    int err = 0;

    /* Assume not found. */
    *found_out = 0;

    /* Open /proc/modules. */
    if ((file = fopen(PATH_PROC_MODULES, "r")) == NULL) {
        return errno;
    }

    for (;;) {
        /* Read the next line. */
        if (fgets(buffer, sizeof(buffer), file) == NULL) {
            if (ferror(file)) {
                err = errno;
            }

            break;
        }

        /* Extract the module name. */
        point = buffer;
        module_name = strsep(&point, ws);
        if (strcmp(module_name, name) == 0) {
            *found_out = 1;
            break;
        }
    }

    fclose(file);
    return err;
}

/* Load a module. */
static int
module_load(const char *name)
{
    int err;
    int found;
    char canon[BUFFER_SIZE];

    /* Canonicalize the module name. */
    module_canonical_name(name, canon, sizeof(canon));

    /* See if the module is already present. */
    if ((err = module_is_loaded(canon, &found)) != 0) {
        return err;
    }

    /* Bail if it is. */
    if (found) {
        return 0;
    }

    /* Make a note. */
    xlog(LOG_DEBUG, "loading module %s", canon);

    /* Invoke the program. */
    return invokev(path_modprobe,
                   "modprobe", MODPROBE_ARGS, canon,
                   NULL);
}

/* Load each of the modules in a comma-separated list. */
static int
modules_load(const char *list)
{
    char buffer[BUFFER_SIZE];
    const char *mark = list;
    const char *point;
    int err;

    for (point = list; *point != 0; point++) {
        if (*point == ',') {
            memcpy(buffer, mark, point - mark);
            buffer[point - mark] = 0;
            if ((err = module_load(buffer)) != 0) {
                return err;
            }

            mark = point + 1;
        }
    }

    return module_load(mark);
}

/* Unload a module. */
static int
module_remove(const char *name)
{
    int err;
    int found;
    char canon[BUFFER_SIZE];

    /* Canonicalize the module name. */
    module_canonical_name(name, canon, sizeof(canon));

    /* Make sure the module is present. */
    if ((err = module_is_loaded(canon, &found)) != 0) {
        return err;
    }

    /* If not, we're done. */
    if (!found) {
        return 0;
    }

    /* Make a note. */
    xlog(LOG_DEBUG, "removing module %s", canon);

    /* Unload the module. */
    if (delete_module(canon, 0) < 0) {
        return errno;
    }

    return 0;
}

/* Removes each of the modules in a comma-separated list. */
static int
modules_remove(const char *list)
{
    char buffer[BUFFER_SIZE];
    const char *mark = list + strlen(list);
    const char *point;
    int err;

    for (point = mark; point >= list; point--) {
        if (*point == ',') {
            memcpy(buffer, point + 1, mark - point - 1);
            buffer[mark - point - 1] = 0;
            if ((err = module_remove(buffer)) != 0) {
                return err;
            }

            mark = point;
        }
    }

    memcpy(buffer, list, mark - list);
    buffer[mark - list] = 0;
    return module_remove(buffer);
}

/* Parse a PVC ATM address of te form [itf.]vpi.vci. */
static int
parse_pvc(const char *const string, struct sockaddr *addr, size_t addrlen)
{
    struct sockaddr_atmpvc name;
    const char *mark = NULL;
    const char *point;
    char *p;
    long values[3];
    pvc_state_t state = ST_START;
    int i, ch;

    i = 0;
    for (point = string; state != ST_END; point++) {
        ch = *point;
        switch (state) {
            case ST_START:
                if (i >= 3) {
                    return EINVAL;
                }

                if (ch == '0') {
                    state = ST_ZERO;
                } else if (isdigit(ch)) {
                    mark = point;
                    state = ST_NUM;
                } else {
                    return EINVAL;
                }
                break;

            case ST_NUM:
                if (ch == '.' || ch == 0) {
                    if ((values[i++] = strtol(mark, &p, 0)) == 0) {
                        return errno;
                    }

                    if (p != point) {
                        return EINVAL;
                    }

                    state = (ch == 0) ? ST_END : ST_START;
                } else if (!isdigit(ch)) {
                    return EINVAL;
                }

                break;

            case ST_ZERO:
                if (ch != '.' && ch != 0) {
                    return EINVAL;
                }

                values[i++] = 0;
                state = ST_START;
                break;

            default:
                abort();
        }
    }

    /* Check the number of numbers found and rearrange. */
    if (state != ST_END || i < 2) {
        return EINVAL;
    } else if (i == 2) {
        values[2] = values[1];
        values[1] = values[0];
        values[0] = -1;
    } else if (i > 3) {
        return EINVAL;
    }

    /* Check ranges. */
    if (values[0] > SHRT_MAX || 
        values[1] > ATM_MAX_VPI ||
        values[2] > ATM_MAX_VCI) {
        return ERANGE;
    }

    /* Construct a name. */
    name.sap_family = AF_ATMPVC;
    name.sap_addr.itf = values[0];
    name.sap_addr.vpi = values[1];
    name.sap_addr.vci = values[2];

    /* Copy it to the result. */
    memcpy(addr, &name, addrlen < sizeof(name) ? addrlen : sizeof(name));
    return 0;
}

/* Initialize an RFC2684 bridge struct. */
static void
br2684_init(struct br2684 *bridge)
{
    bridge->sock = -1;
    bridge->ifname[0] = 0;
    bridge->session_id = 0;
}

/* Assign virtual circuit numbers. */
static int
br2684_assign_vc(struct br2684 *bridge, int index, int encap,
                 struct sockaddr *addr, socklen_t addrlen)
{
    struct atm_newif_br2684 req;
    struct atm_qos qos;
    struct atm_backend_br2684 back;
    int bufsize = 8192;
    int sock, err;

    /* Construct the interface name. */
    snprintf(bridge->ifname, sizeof(bridge->ifname), "nas%d", index);
    xlog(LOG_DEBUG, "assigning virtual circuit %d.%d.%d to %s",
         ((struct sockaddr_atmpvc *)addr)->sap_addr.itf,
         ((struct sockaddr_atmpvc *)addr)->sap_addr.vpi,
         ((struct sockaddr_atmpvc *)addr)->sap_addr.vci,
         bridge->ifname);


    /* Create a socket to talk to the ATM subsystem.  Note that when
     * this socket is closed, the bridge disappears, but the interface
     * doesn't. */
    if ((sock = socket(PF_ATMPVC, SOCK_DGRAM, ATM_AAL5)) < 0) {
        err = errno;
        xlog(LOG_DEBUG, "socket failed: %s", strerror(err));
        return err;
    }

    /* Create the interface for pppd to talk to. */
    memset(&req, 0, sizeof(req));
    req.backend_num = ATM_BACKEND_BR2684;
    req.media = BR2684_MEDIA_ETHERNET;
    req.mtu = 1500;
    strncpy(req.ifname, bridge->ifname, sizeof(req.ifname));

    if ((ioctl(sock, ATM_NEWBACKENDIF, &req)) != 0) {
        if (errno != EEXIST) {
            err = errno;
            xlog(LOG_DEBUG, "unable to create ATM backend interface: %s",
                 strerror(err));
            close(sock);
            return err;
        }
    }

    /* Set the size of the send buffer. */
    if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF,
                   &bufsize, sizeof(bufsize)) < 0) {
        err = errno;
        xlog(LOG_DEBUG, "setsockopt (SO_SNDBUF) failed: %s", strerror(errno));
        close(sock);
        return err;
    }

    /* Set the quality of service. */
    memset(&qos, 0, sizeof(qos));
    qos.txtp.traffic_class = ATM_UBR;
    qos.txtp.pcr = ATM_MAX_PCR;
    qos.txtp.max_sdu = 1524;
    qos.rxtp.traffic_class = ATM_UBR;
    qos.rxtp.pcr = ATM_MAX_PCR;
    qos.rxtp.max_sdu = 1524;
    qos.aal = ATM_AAL5;

    if (setsockopt(sock, SOL_ATM, SO_ATMQOS, &qos, sizeof(qos)) < 0) {
        err = errno;
        xlog(LOG_DEBUG, "setsockopt (SO_ATMQOS) failed: %s",
             strerror(err));
        close(sock);
        return err;
    }

    /* Connect it to the right virtual circuit. */
    if (connect(sock, addr, addrlen) < 0) {
        err = errno;
        xlog(LOG_DEBUG, "connect failed: %s", strerror(err));
        close(sock);
        return err;
    }

    /* Configure the back-end information. */
    memset(&back, 0, sizeof(back));
    back.backend_num = ATM_BACKEND_BR2684;
    back.ifspec.method = BR2684_FIND_BYIFNAME;
    strncpy(back.ifspec.spec.ifname, bridge->ifname,
            sizeof(back.ifspec.spec.ifname));
    back.fcs_in = BR2684_FCSIN_NO;
    back.fcs_out = BR2684_FCSOUT_NO;
    back.encaps = encap;

    if (ioctl(sock, ATM_SETBACKEND, &back) != 0) {
        err = errno;
        xlog(LOG_DEBUG, "ioctl (ATM_SETBACKEND) failed: %s",
             strerror(err));
        close(sock);
        return err;
    }

    /* Bring up the interface. */
    err = invokev(PATH_IFCONFIG, "ifconfig", bridge->ifname, "up", NULL);
    if (err != 0) {
        xlog(LOG_DEBUG, "ifconfig failed: %d", err);
        close(sock);
        return err;
    }

    /* Record the socket. */
    bridge->sock = sock;
    xlog(LOG_DEBUG, "virtual circuit %s assigned", bridge->ifname);
    return 0;
}

/* Unassign a virtual circuit. */
static int
br2684_unassign_vc(struct br2684 *bridge)
{
    int err, sock;

    xlog(LOG_DEBUG, "unassigning virtual circuit %s", bridge->ifname);

    /* Take responsibility for closing the socket. */
    sock = bridge->sock;
    bridge->sock = -1;

    /* Bring down the interface. */
    err = invokev(PATH_IFCONFIG, "ifconfig", bridge->ifname, "down", NULL);
    if (err != 0) {
        xlog(LOG_DEBUG, "ifconfig failed: %d", err);
        close(sock);
        return err;
    }

    /* Close the socket to disable the ATM back-end. */
    if (close(sock) < 0) {
        err = errno;
        xlog(LOG_DEBUG, "unable to close socket: %s\n",
             strerror(err));
        return err;
    }

    return 0;
}

/* Create a packet capture socket.  If a PPPoE link dies without the
 * client properly closing the connection then the server may refuse
 * to create a new link until the old session times out.  To expedite
 * the process, we listen for PPPoE session packets sent to us and
 * pluck the remote MAC address and the session id out of the first
 * one we receive and use those to generate a PPPoE Active Discovery
 * Terminate (PADT) packet if pppd dies with an authentication
 * error. */
static int
make_pcap_socket(const char *ifname, short protocol, int *sock_out)
{
    int err, sock;
    struct ifreq ifr;
    struct sockaddr_ll name;
    socklen_t len;

    /* Create a raw socket. */
    sock = socket(PF_PACKET, SOCK_RAW, protocol);
    if (sock < 0) {
	return errno;
    }

    /* Look up the index of the interface. */
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
    if (ioctl(sock, SIOCGIFINDEX, &ifr) < 0) {
	err = errno;
	close(sock);
	return err;
    }

    /* Bind the socket to the interface. */
    memset(&name, 0, sizeof(name));
    name.sll_family = AF_PACKET;
    name.sll_ifindex = ifr.ifr_ifindex;
    name.sll_protocol = protocol;
    if (bind(sock, (struct sockaddr *)&name, sizeof(name)) < 0) {
	err = errno;
	close(sock);
	return err;
    }

    /* Check for errors. */
    len = sizeof(err);
    if (getsockopt(sock, SOL_SOCKET, SO_ERROR, &err, &len) < 0) {
	err = errno;
	close(sock);
	return err;
    }

    if (err) {
	close(sock);
	return err;
    }

    /* Return the new socket. */
    *sock_out = sock;
    return 0;
}

/* Get the MAC address of the pcap socket's interface. */
static int
pcap_ifhwaddr(int sock, const char *ifname, unsigned char *addr_out)
{
    struct ifreq ifr;

    /* Look up the MAC address. */
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
    if (ioctl(sock, SIOCGIFHWADDR, &ifr) < 0) {
        return errno;
    }

    /* Return it. */
    memcpy(addr_out, ifr.ifr_hwaddr.sa_data, IFHWADDRLEN);
    return 0;
}

/* Read PPPoE session packet. */
static int
pcap_recv(int sock, struct br2684 *bridge)
{
    struct lcp_pkt *pkt;
    char buffer[2048];
    ssize_t len;

    /* Read a packet. */
    len = recv(sock, buffer, sizeof(buffer), MSG_TRUNC);
    if (len < 0) {
        return errno;
    }

    /* Ignore short packets. */
    if (len < sizeof(struct lcp_pkt)) {
        return 0;
    }

    /* Treat the buffer as a packet. */
    pkt = (struct lcp_pkt *)buffer;

    /* Sanity check. */
    ASSERT(pkt->eth.h_proto == htons(ETH_P_PPP_SES));

    /* Ignore anything wihch isn't an LCP echo request or reply with
     * our mac address as the destination. */
    if (pkt->proto != htons(PPP_LCP) ||
	 pkt->lcp.code != LCP_ECHO_RQST ||
	memcmp(bridge->mac_local, pkt->eth.h_dest, IFHWADDRLEN) != 0) {
	return 0;
    }

    /* Record the sender's address. */
    memcpy(bridge->mac_peer, pkt->eth.h_source, IFHWADDRLEN);

    /* And the session id. */
    bridge->session_id = ntohs(pkt->pppoe.sid);
    return 0;
}


/* Send a PADT packet. */
static int
send_padt(struct br2684 *bridge)
{
    int err, sock;
    struct padt_pkt packet;

    /* Construct a PADT packet. */
    memset(&packet, 0, sizeof(packet));
    memcpy(&packet.eth.h_dest, bridge->mac_peer, ETH_ALEN);
    memcpy(&packet.eth.h_source, bridge->mac_local, ETH_ALEN);
    packet.eth.h_proto = htons(ETH_P_PPP_DISC);
    packet.pppoe.ver = 1;
    packet.pppoe.type = 1;
    packet.pppoe.code = PADT_CODE;
    packet.pppoe.sid = htons(bridge->session_id);
    packet.pppoe.length = 0;

    /* Create a raw socket. */
    err = make_pcap_socket(bridge->ifname, htons(ETH_P_PPP_DISC), &sock);
    if (err) {
        return err;
    }

    /* Send the packet. */
    if (send(sock, &packet, sizeof(packet), 0) < 0) {
        err = errno;
        close(sock);
        return err;
    }

    /* Close the socket. */
    if (close(sock) < 0) {
        return errno;
    }

    return 0;
}

/* Locates the specified device in /proc/net/atm/devices.  An itf less
 * than zero is treated as a wildcard, as is a NULL mac_addr.  If
 * neither is specified then the first device in the list will be
 * used. */
static int
find_device(int itf, const char *mac_addr,
            int *itf_out, char *path_out, size_t path_size)
{
    FILE *file;
    int iface;
    char typestr[9];
    char mac[16];
    char buffer[256];
    int found = 0;
    int err = 0;

    /* Open the file. */
    if (itf < 0) {
        xlog(LOG_DEBUG, "looking for any interface in %s",
             PATH_PROC_DEVICES);
    } else {
        xlog(LOG_DEBUG, "looking for interface %d in %s",
             itf, PATH_PROC_DEVICES);
    }

    if ((file = fopen(PATH_PROC_DEVICES, "r")) == NULL) {
        err = errno;
        xlog(LOG_DEBUG, "unable to open file %s: %s",
             PATH_PROC_DEVICES, strerror(err));
        return err;
    }

    /* Skip the first line. */
    if (fgets(buffer, sizeof(buffer), file) == NULL) {
        if (ferror(file)) {
            err = errno;
            xlog(LOG_DEBUG, "error reading from file %s: %s",
                 PATH_PROC_DEVICES, strerror(err));
            fclose(file);
            return err;
        }
    }

    /* Scan through the devices. */
    while (fscanf(file, "%3d %8s %15s %255[^\n]\n",
                  &iface, typestr, mac, buffer) == 4) {
        /* Compare the ATM interface numbers. */
        if (itf != -1 && iface != itf) {
            xlog(LOG_DEBUG, "looking for interface %d, found %d",
                 itf, iface);
            continue;
        }

        /* Compare the mac address. */
        if (mac_addr != NULL) {
            if (strcmp(mac_addr, mac) != 0) {
                xlog(LOG_DEBUG, "looking for MAC address %s, found %s",
                     mac_addr, mac);
                continue;
            }
        }

        /* Otherwise we have a match. */
        *itf_out = iface;
        snprintf(path_out, path_size, PATH_PROC_ATM "/%s:%d", typestr, iface);
        xlog(LOG_DEBUG, "found interface %d, MAC address %s",
             iface, mac);
        found = 1;
        break;
    }

    /* Check if an error occurred. */
    if (ferror(file)) {
        err = errno;
        xlog(LOG_DEBUG, "error while reading file %s: %s\n",
             PATH_PROC_DEVICES, strerror(err));
    }

    /* Close the file. */
    if (fclose(file) == EOF) {
        err = err ? err : errno;
        xlog(LOG_DEBUG, "error closing file %s: %s\n",
             PATH_PROC_DEVICES, strerror(errno));
    }

    if (!found) {
        err = err ? err : ENOENT;
    }
    return err;
}

/* Run pppd with the given args. */
static int
run_pppd(int argc, char **argv, pid_t *pid_out)
{
    int err;
    char buffer[BUFFER_SIZE];
    int offset = 0;
    char **args;
    int i;
    long fd_count;
    int fd;
    pid_t pid;

    /* Allocate an array for args. */
    if ((args = alloca((argc + 3) * sizeof(char *))) == NULL) {
        return errno;
    }

    /* Construct pppd's argv. */
    args[0] = "pppd";
    args[1] = "nodetach";
    for (i = 0; i < argc; i++) {
        args[i + 2] = argv[i];
    }
    args[argc + 2] = NULL;

    /* Copy the args into buffer for logging. */
    offset = 0;
    buffer[0] = '\0';
    for (i = 0; args[i] != NULL && offset < sizeof(buffer); i++) {
        offset += snprintf(buffer + offset, sizeof(buffer) - offset,
                           "%s ", args[i]);
    }

    /* Fork a child process. */
    if ((pid = fork()) == (pid_t)-1) {
        err = errno;
        free(args);
        *pid_out = pid;
        return err;
    }

    /* If we're the parent then return. */
    if (pid != 0) {
        *pid_out = pid;
        xlog(LOG_DEBUG, "spawned child pid=%u", *pid_out);
        xlog(LOG_DEBUG, "argv is: %s", buffer);
        return 0;
    }

    /* We're the child.  Close off all file descriptors. */
    fd_count = sysconf(_SC_OPEN_MAX);
    for (fd = 0; fd < fd_count; fd++) {
        close(fd);
    }

    /* Replace stdin, stdout and stderr with /dev/null. */
    if (open(_PATH_DEVNULL, O_RDONLY) < 0 ||
        (fd = open(_PATH_DEVNULL, O_WRONLY)) < 0 ||
        dup2(fd, 2) < 0) {
        xlog(LOG_ERR, "error: unable to set up stdin, stdout and stderr: %s",
             xstrerror(errno));
        exit(EXIT_FATAL_ERROR);
    }

    /* Run pppd. */
    execv(PATH_PPPD, args);
    syslog(LOG_ERR, "error: execv %s failed: %s", PATH_PPPD, xstrerror(errno));
    exit(EXIT_FATAL_ERROR);
}

/* Initialize the fsm. */
static void
fsm_init(struct fsm *self)
{
    memset(self, 0, sizeof(struct fsm));
    self->state = FSM_START;
    self->exit_on_error = 1;
    self->daemonize = 1;
    self->atm_itf = -1;
    self->status = STATUS_NO_MODEM;
    self->malformed_count = -1;
    self->pppd_pid = (pid_t)-1;
    self->conntype = CT_PPPoA;
    self->encap = BR2684_ENCAPS_LLC;
    br2684_init(&self->bridge);
    self->pcap_sock = -1;
}

/* Tell the ancestor process to exit. */
static void
fsm_detach(struct fsm *self)
{
    int fd;

    /* Note what we're doing. */
    xlog(LOG_DEBUG, "Detaching");

    /* Write the status to stdout. */
    fputc(0, stdout);
    fflush(stdout);

    /* Open /dev/null. */
    if ((fd = open(_PATH_DEVNULL, O_WRONLY)) < 0) {
        xlog(LOG_ERR, "error: unable to open " _PATH_DEVNULL ": %s\n",
             xstrerror(errno));
        exit(1);
    }

    /* Close stdout and stderr. */
    close(STDOUT_FILENO);
    close(STDERR_FILENO);

    /* Redirect them to /dev/null. */
    dup2(fd, STDOUT_FILENO);
    dup2(fd, STDERR_FILENO);

    /* Stop logging to stderr. */
    log_to_stderr = 0;

    /* We've detached. */
    self->daemonize = 0;

    /* Don't exit when faced with a recoverable error. */
    self->exit_on_error = 0;
}

/* Cope with an error. */
static void
fsm_cleanup(struct fsm *self, int flags)
{
    int err;

    /* Disable the RFC 2684 bridge. */
    if ((flags & CU_BRIDGE) && self->conntype == CT_PPPoE) {
	/* Close the packet capture socket if it's open. */
	if (self->pcap_sock != -1) {
	    if (close(self->pcap_sock) < 0) {
		xlog(LOG_ERR, "error: unable to close packet capture socket: %s\n",
		     xstrerror(errno));
	    }
	    self->pcap_sock = -1;
	}

	/* The session id is no longer valid. */
	self->bridge.session_id = 0;

	/* Close the bridge socket. */
	if (self->bridge.sock != -1) {
	    err = br2684_unassign_vc(&self->bridge);
	    if (err) {
		xlog(LOG_ERR, "error: unable to unassign virtual circuit: %s",
		     xstrerror(err));
	    }
	}
    }

    /* Unload the modules. */
    if ((flags & CU_MODULES) && self->mod_list) {
	err = modules_remove(self->mod_list);
	if (err) {
	    xlog(LOG_ERR, "error: unable to remove modules: %s",
		 xstrerror(err));
	}
    }
}

/* Exit on errors. */
static void
fsm_nonfatal_error(struct fsm *self)
{
    if (!self->exit_on_error) {
	return;
    }

    /* This cannot be called if pppd is running. */
    ASSERT(self->pppd_pid == (pid_t)-1);

    /* Clean up the bridge and unload modules. */
    fsm_cleanup(self, CU_MODULES | CU_BRIDGE);

    /* Let the foreground process know we've had an error. */
    if (self->daemonize) {
	fputc(1, stdout);
	fflush(stdout);
    }

    /* And exit. */
    exit(1);
}

static pid_t
fsm_reap_pppd(struct fsm *self, int *code_out)
{
    pid_t pid;
    int status;

    /* A child process has died.  Wait for it. */
    pid = waitpid(self->pppd_pid, &status, 0);
    if (pid == (pid_t)-1) {
	xlog(LOG_ERR, "error: waitpid failed: %s", xstrerror(errno));
	exit(1);
    }

    /* Ignore it if it wasn't pppd. */
    if (pid == 0) {
	return pid;
    }
    ASSERT(pid == self->pppd_pid);
    self->pppd_pid = (pid_t)-1;

    if (WIFEXITED(status)) {
	*code_out = WEXITSTATUS(status);
	xlog(LOG_INFO, "pppd exited with code %d", *code_out);
    } else if (WIFSIGNALED(status)) {
	xlog(LOG_INFO, "pppd exited on signal %d", WTERMSIG(status));
	*code_out = -1;
    } else {
	xlog(LOG_INFO, "pppd died with status %d", status);
	*code_out = 0;
    }

    return pid;
}

static void
fsm_start(struct fsm *self, int input)
{
    int err;

    if (input == SIGALRM) {
	/* If we have a list of modules then load them. */
	if (self->mod_list) {
	    err = modules_load(self->mod_list);
	    if (err) {
		xlog(LOG_ERR, "modprobe %s failed: %s",
		     self->mod_list, xstrerror(err));
		exit(1);
	    }
	}

	/* Record the time when we should give up on finding the modem
	 * status file in /proc. */
	WAKESET(MODEM_INIT_TIMEOUT, &self->expire);
	SET_STATE(self->state, FSM_FIND_STATUS);
	WAKESETNOW(&self->wake);
	return;
    }

    xlog(LOG_ERR, "internal error: unhandled input in %s: %d",
	 __FUNCTION__, input);
    exit(1);
}

/* Look for the device status file. */
static void
fsm_find_status(struct fsm *self, int input)
{
    int err, itf = -1;
    struct timeval now;

    if (input == SIGALRM) {
        err = find_device(self->atm_itf, self->mac_addr, &itf,
                          self->path_modem_status,
                          sizeof(self->path_modem_status));

        if (err == ENOENT) {
            /* See if we've waited long enough. */
            xgettimeofday(&now);
            if (timercmp(&now, &self->expire, >)) {
                xlog(LOG_ERR, "error: unable to locate ATM device: %s",
                     xstrerror(err));

                /* Unload the modules. */
                fsm_cleanup(self, CU_MODULES);

                /* Die if we're doing that. */
		fsm_nonfatal_error(self);

                /* Otherwise start again. */
                SET_STATE(self->state, FSM_START);
                WAKESET(SHORT_PAUSE, &self->wake);
                return;
            }

            /* Not yet.  Wait a bit before trying again. */
            WAKESET(SHORT_PAUSE, &self->wake);
            return;
        }

        if (err) {
            /* Any other error is his fatal. */
            xlog(LOG_ERR, "error: unable to read ATM device list: %s",
                 xstrerror(err));
            fsm_cleanup(self, CU_MODULES);
            exit(1);
        }

        /* Found the device. */
        xlog(LOG_INFO, "reading modem status from %s",
             self->path_modem_status);

        /* If we're doing PPPoE then record the interface number in
         * the ATM address. */
        if (self->conntype == CT_PPPoE) {
            self->atm_addr.sap_addr.itf = itf;
        }

        /* Reset the malformed count. */
        self->malformed_count = 0;

        /* Determine when to give up waiting for the ATM link. */
        WAKESET(MODEM_LINEUP_TIMEOUT, &self->expire);

        /* And wait of it. */
        SET_STATE(self->state, FSM_CARRIER);
        WAKESET(SHORT_PAUSE, &self->wake);
        return;
    }

    xlog(LOG_ERR, "internal error: unhandled input in %s: %d",
         __FUNCTION__, input);
    exit(1);
}

static void
fsm_carrier(struct fsm *self, int input)
{
    int err;
    status_t status = STATUS_NO_MODEM;
    struct timeval now;

    if (input == SIGALRM) {
        err = modem_get_status(self->path_modem_status, &status);
        if (err) {
            /* Don't panic immediately if we can't read the file, but
             * keep track of the number of times this fails in case
             * the proc file format changes on us. */
            if (++self->malformed_count < MAX_MALFORMED) {
                xlog(LOG_WARNING, "warning: unable to determine modem status "
                     "from file %s: %s", self->path_modem_status,
                     xstrerror(err));
                WAKESET(SHORT_PAUSE, &self->wake);
                return;
            }

            /* We've failed to read the modem status several times in
             * a row.  It's time to give up. */
            xlog(LOG_ERR, "error: unable to determine modem status from "
                 "file %s: %s", self->path_modem_status, xstrerror(err));
            fsm_cleanup(self, CU_MODULES);

            /* Die if exiting on errors. */
	    fsm_nonfatal_error(self);

            /* Otherwise try again. */
            SET_STATE(self->state, FSM_START);
            WAKESET(SHORT_PAUSE, &self->wake);
            return;
        }

        /* Note changes in the modem state. */
        if (self->status != status) {
            xlog(LOG_INFO, "modem status: %s", status_names[status]);
            self->status = status;
        }

        /* Is the line still not up? */
        if (status != STATUS_LINE_UP) {
            /* Is it time to give up? */
            xgettimeofday(&now);
            if (timercmp(&now, &self->expire, >)) {
                xlog(LOG_ERR, "error: ATM link not found in a timely fashion");

                /* Unload the modules. */
                fsm_cleanup(self, CU_MODULES);

                /* Die if exiting on errors. */
		fsm_nonfatal_error(self);

                /* Otherwise start again shortly. */
                SET_STATE(self->state, FSM_START);
                WAKESET(SHORT_PAUSE, &self->wake);
                return;
            }

            /* Check again later. */
            WAKESET(SHORT_PAUSE, &self->wake);
            return;
        }

        /* Line is up!  If we're doing PPPoE then build an RFC 2684
         * bridge. */
        if (self->conntype == CT_PPPoE) {
            err = br2684_assign_vc(&self->bridge,
                                   self->atm_addr.sap_addr.itf,
                                   self->encap,
                                   (struct sockaddr *)&self->atm_addr,
                                   sizeof(self->atm_addr));
            if (err) {
                xlog(LOG_ERR, "error: unable to assign ATM back-end: %s",
                     xstrerror(err));

                /* Unload the modules. */
                fsm_cleanup(self, CU_MODULES);

                /* Die if exiting on errors. */
		fsm_nonfatal_error(self);

                /* Otherwise start again. */
                SET_STATE(self->state, FSM_START);
                WAKESET(SHORT_PAUSE, &self->wake);
                return;
            }

            /* Create a packet capture socket so that we can listen
             * for LCP echo request packets from the peer. */
            err = make_pcap_socket(self->bridge.ifname,
                                   htons(ETH_P_PPP_SES),
                                   &self->pcap_sock);
            if (err) {
                xlog(LOG_WARNING, "warning: unable to create packet capture "
                     "socket: %s", xstrerror(err));
                self->pcap_sock = -1;
            } else {
                /* Look up the local MAC address. */
                err = pcap_ifhwaddr(self->pcap_sock, self->bridge.ifname,
                                    self->bridge.mac_local);
                if (err) {
                    xlog(LOG_WARNING, "warning: unable to determine local "
                         "hardware address: %s", xstrerror(err));
                    close(self->pcap_sock);
                    self->pcap_sock = -1;
                }
            }
        }

        /* Fire up the ppp daemon after a suitable pause. */
        SET_STATE(self->state, FSM_PPPD);
        WAKESET(LINE_UP_SETTLE_PAUSE, &self->wake);
        return;
    }

    xlog(LOG_ERR, "internal error: unhandled input in %s: %d",
         __FUNCTION__, input);
    exit(1);
}

static void
fsm_pppd(struct fsm *self, int input)
{
    int err;

    /* If we receive an LCP Echo-Request from our peer *before* we
     * start pppd then we know that there's an unterminated session
     * still going.  Send a PADT packet to terminate it and reset the
     * timeout. */
    if (input == SIGCONT) {
	ASSERT(self->bridge.session_id != 0);
	err = send_padt(&self->bridge);
	self->bridge.session_id = 0;

	if (err) {
	    xlog(LOG_ERR, "error: failed to send PADT packet: %s",
		 xstrerror(errno));

	    /* If we exit on errors then do so. */
	    fsm_nonfatal_error(self);

	    /* Otherwise pretend that nothing happened. */
	    return;
	}

	/* Reset the timeout. */
	WAKESET(LINE_UP_SETTLE_PAUSE, &self->wake);
	return;
    }

    if (input == SIGALRM) {
	/* Fire up pppd. */
	err = run_pppd(self->pppd_argc, self->pppd_argv, &self->pppd_pid);
	if (err) {
	    xlog(LOG_ERR, "error: unable to spawn pppd: %s",
		 xstrerror(err));

	    /* Die if we exit on errors. */
	    fsm_nonfatal_error(self);

	    /* Otherwise return to the carrier state. */
	    fsm_cleanup(self, CU_BRIDGE);
	    SET_STATE(self->state, FSM_CARRIER);
	    WAKESET(PPPD_RESTART_TIMEOUT, &self->wake);
	    return;
	}

	/* If we have a packet capture socket then we're doing PPPoE
	 * and are listening for LCP Echo-Request packets from the
	 * peer in order to detect an unterminated session.  We'll
	 * wait in the FSM_SESSION state for a bit. */
	if (self->pcap_sock >= 0) {
	    ASSERT(self->bridge.session_id == 0);
	    SET_STATE(self->state, FSM_SESSION);
	    WAKESET(PPPoE_SESS_TIMEOUT, &self->wake);
	    return;
	}

	/* Otherwise now is a good time to detach. */
	if (self->daemonize && !self->detach_on_signal) {
	    fsm_detach(self);
	}

	/* Go to the FSM_UP state. */
	xlog(LOG_INFO, "Running");
	SET_STATE(self->state, FSM_UP);
	WAKESETNONE(&self->wake);
	return;
    }

    xlog(LOG_ERR, "internal error: unhandled input in %s: %d",
         __FUNCTION__, input);
    exit(1);
}

static void
fsm_session(struct fsm *self, int input)
{
    int code;
    pid_t pid;

    if (input == SIGCONT) {
	/* We've received an LCP Echo-Request packet from the peer.
	 * Move to the FSM_UP state, and now is a good time to detach
	 * if we're doing that. */
	if (self->daemonize && !self->detach_on_signal) {
	    fsm_detach(self);
	}

	/* The packet capture socket should be closed. */
	ASSERT(self->pcap_sock == -1);

	/* Go to the FSM_UP state. */
	xlog(LOG_INFO, "Running");
	SET_STATE(self->state, FSM_UP);
	WAKESETNONE(&self->wake);
	return;
    }

    if (input == SIGALRM) {
	/* We've given up on waiting for the peer to send us an LCP
	 * Echo-Request packet.  If we've managed to stay connected
	 * this long then it's reasonable to assume that we aren't
	 * suffering from an unterminated PPPoE session, and it's a
	 * reasonably good time to detach. */
	if (self->daemonize && !self->detach_on_signal) {
	    fsm_detach(self);
	}

	/* Close the packet capture socket. */
	ASSERT(self->pcap_sock != -1);
	if (close(self->pcap_sock) < 0) {
	    xlog(LOG_ERR, "error: unable to close packet capture socket: %s\n",
		 xstrerror(errno));
	}
	self->pcap_sock = -1;

	/* Go to the FSM_UP state. */
	xlog(LOG_INFO, "Running");
	SET_STATE(self->state, FSM_UP);
	WAKESETNONE(&self->wake);
	return;
    }

    if (input == SIGCHLD) {
	/* A child process has died.  Reap it. */
	pid = fsm_reap_pppd(self, &code);

	/* Ignore it if it wasn't pppd. */
	if (pid == 0) {
	    return;
	}

	/* If we're still listening for LCP Echo-Request packets
	 * and pppd died with an error which suggests that another
	 * PPPoE session is active then go to the FSM_PADT and
	 * continue waiting for an Echo-Request so that we can
	 * terminate the old session. */
	if (self->pcap_sock >= 0 && (1 << code) & PADT_EXIT_CODES) {
	    ASSERT(self->bridge.session_id == 0);
	    SET_STATE(self->state, FSM_PADT);
	    /* The existing timeout is fine. */
	    return;
	}

	/* If we're exiting on errors then do so now. */
	fsm_nonfatal_error(self);

	/* Tear down the bridge. */
	fsm_cleanup(self, CU_BRIDGE);

	/* Wait a bit and try again. */
	SET_STATE(self->state, FSM_CARRIER);
	WAKESET(PPPD_RESTART_TIMEOUT, &self->wake);
	return;
    }

    xlog(LOG_ERR, "internal error: unhandled input in %s: %d",
         __FUNCTION__, input);
    exit(1);
}

static void
fsm_up(struct fsm *self, int input)
{
    int err, code;
    pid_t pid;

    if (input == SIGCHLD) {
	/* A child process has died.  Reap it. */
	pid = fsm_reap_pppd(self, &code);

	/* We only care if the dead child was pppd. */
	if (pid == 0) {
	    return;
	}

	/* If pppd has died with an error which suggests that another
	 * PPPoE session is active and we have a session-id from an
	 * LCP Echo-Request packet from the peer then send a PADT
	 * packet to terminate that session. */
	if (self->bridge.session_id && (1 << code) & PADT_EXIT_CODES) {
	    /* Send a PADT packet. */
	    xlog(LOG_DEBUG, "sending PADT packet for session 0x%hx",
		 self->bridge.session_id);
	    err = send_padt(&self->bridge);
	    self->bridge.session_id = 0;
	    if (!err) {
		/* Try running pppd again. */
		SET_STATE(self->state, FSM_PPPD);
		WAKESETNOW(&self->wake);
		return;
	    }
	}

	/* Die if exiting on errors. */
	fsm_nonfatal_error(self);

	/* Tear down the bridge. */
	fsm_cleanup(self, CU_BRIDGE);
	
	/* Wait a bit and try again. */
	SET_STATE(self->state, FSM_CARRIER);
	WAKESET(PPPD_RESTART_TIMEOUT, &self->wake);
	return;
    }

    xlog(LOG_ERR, "internal error: unhandled input in %s: %d",
         __FUNCTION__, input);
    exit(1);
}

static void
fsm_padt(struct fsm *self, int input)
{
    int err;
    unsigned short session_id;

    if (input == SIGALRM) {
	/* Timed out waiting for an LCP Echo-Request packet.  Assume
	 * that we just missed the last one and try again.  Tear down
	 * the bridge.*/
	fsm_cleanup(self, CU_BRIDGE);

	/* We've already waited a bit, so try again now. */
	SET_STATE(self->state, FSM_CARRIER);
	WAKESETNOW(&self->wake);
	return;
    }

    if (input == SIGCONT) {
	/* We've received a packet so we must have a session-id.  Send
	 * a PADT packet to terminate it. */
	ASSERT(self->bridge.session_id);
	session_id = self->bridge.session_id;
	err = send_padt(&self->bridge);
	self->bridge.session_id = 0;

	if (err) {
	    xlog(LOG_ERR, "error: failed to send PADT packet: %s",
		 xstrerror(err));

	    /* Die if we exit on errors. */
	    fsm_nonfatal_error(self);

	    /* Otherwise wait a bit and try again. */
	    SET_STATE(self->state, FSM_CARRIER);
	    WAKESET(PPPD_RESTART_TIMEOUT, &self->wake);
	    return;
	}

	xlog(LOG_DEBUG, "sent PADT packet for session 0x%hx", session_id);

	/* Try running pppd again. */
	SET_STATE(self->state, FSM_PPPD);
	WAKESETNOW(&self->wake);
	return;
    }

    xlog(LOG_ERR, "internal error: unhandled input in %s: %d",
         __FUNCTION__, input);
    exit(1);
}

static void
fsm_reap(struct fsm *self, int input)
{
    pid_t pid;
    int code;

    if (input == SIGCHLD) {
	pid = fsm_reap_pppd(self, &code);

	/* We only care if the child was pppd. */
	if (pid == 0) {
	    return;
	}

	/* Clean up everything and shut down. */
	fsm_cleanup(self, CU_MODULES | CU_BRIDGE);
	exit(0);
    }

    xlog(LOG_ERR, "internal error: unhandled input in %s: %d",
         __FUNCTION__, input);
    exit(1);
}

/* Clean up on exit. */
static void
clean_up_on_exit()
{
    /* Remove the pid file. */
    unlink(pid_file);
}

/* Clean up shop so that we can be a good daemon. */
static void
daemonize()
{
    pid_t pid;
    FILE *file;
    long fd_count;
    int i, fd, status, pipefds[2];
    sigset_t sigset, oldset;
    ssize_t len;
    char byte;

    /* Defer signals that we intend to forward to the grandchild
     * process until that child is in a position to receive them. */
    sigemptyset(&sigset);
    for (i = 0; forwarded_signals[i]; i++) {
        ASSERT(sigaddset(&sigset, forwarded_signals[i]) == 0);
    }
    ASSERT(sigprocmask(SIG_BLOCK, &sigset, &oldset) == 0);

    /* Open a pipe to communicate with the child. */
    if (pipe(pipefds) < 0) {
        fprintf(stderr, "%s: error: unable to create a pipe: %s\n",
                progname, strerror(errno));
        exit(1);
    }

    /* Spawn a child process. */
    if ((pid = fork()) == (pid_t)-1) {
        fprintf(stderr, "%s: error: unable to spawn child process: %s\n",
                progname, strerror(errno));
        exit(1);
    }

    /* If we're the parent process then wait for the child to
     * detach. */
    if (pid) {
        xlog(LOG_DEBUG, "spawned child pid=%ld", (long)pid);

        /* Close the output end of the pipe. */
        close(pipefds[1]);

        /* Wait for the grandchild to send us its pid. */
        if ((child_pid = read_pid(pipefds[0])) == (pid_t)-1) {
            fprintf(stderr,
                    "%s: error: unable to spawn grandchild process: %s\n",
                    progname, xstrerror(errno));
            exit(1);
        }

        /* Note the grandchild's pid. */
        xlog(LOG_DEBUG, "spawned grandchild pid=%ld", (long)child_pid);

        /* Arrange to forward signals to the grandchild. */
        for (i = 0; forwarded_signals[i]; i++) {
            signal(forwarded_signals[i], forward_signal);
        }

        /* Unblock the signals. */
        if (sigprocmask(SIG_SETMASK, &oldset, NULL) < 0) {
            fprintf(stderr, "%s: error: sigprocmask failed: %s\n",
                    progname, xstrerror(errno));
            exit(1);
        }

        /* Reap the child process. */
        if ((pid = waitpid(pid, &status, 0)) == (pid_t)-1) {
            fprintf(stderr, "%s: error: unable to reap child process: %s\n",
                    progname, xstrerror(errno));
            exit(1);
        }
        xlog(LOG_DEBUG, "reaped child pid=%d", (long)pid);

        /* Wait for the grandchild to detach. */
        if ((len = read(pipefds[0], &byte, 1)) < 0) {
            fprintf(stderr, "%s: error: read failed: %s\n",
                    progname, xstrerror(errno));
            exit(1);
        }

        /* If the pipe was closed without any data then assume that
         * the child died unexpectedly. */
        if (len == 0) {
            exit(1);
        }

        exit(byte);
    }

    /* Replace stdout with the pipe. */
    if (dup2(pipefds[1], STDOUT_FILENO) < 0) {
        xlog(LOG_ERR, "error: dup2 failed: %s\n", xstrerror(errno));
        exit(1);
    }
    
    /* Create a new session and lose the controlling TTY. */
    if (setsid() == (pid_t)-1) {
        xlog(LOG_ERR, "error: unable to set the session id: 0x%x",
             xstrerror(errno));
        exit(1);
    }

    /* Fork again so that we can never acquire a controlling TTY. */
    if ((pid = fork()) == (pid_t)-1) {
        xlog(LOG_ERR, "error: unable to spawn child process: %s",
             xstrerror(errno));
        exit(1);
    }

    /* The parent must die. */
    if (pid != 0) {
        exit(1);
    }

    /* Write the pid to a file. */
    if ((file = fopen(pid_file, "w")) == NULL) {
        xlog(LOG_WARNING, "warning: unable to write pid file %s: %s",
             pid_file, xstrerror(errno));
    } else {
        fprintf(file, "%d\n", getpid());
        fclose(file);
    }

    /* Arrange to clean up when we exit. */
    if (atexit(clean_up_on_exit) < 0) {
        xlog(LOG_ERR, "error: atexit failed: %s", xstrerror(errno));
        exit(1);
    }

    /* Unblock signals. */
    if (sigprocmask(SIG_SETMASK, &oldset, NULL) < 0) {
        xlog(LOG_ERR, "error: sigprocmask failed: %s\n", xstrerror(errno));
        exit(1);
    }

    /* Move to the root directory. */
    if (chdir("/") < 0) {
        xlog(LOG_ERR, "error: unable to change to root directory: %s",
             xstrerror(errno));
        exit(1);
    }

    /* Set the umask to zero. */
    umask(0);

    /* Determine how many file descriptors we're dealing with. */
    fd_count = sysconf(_SC_OPEN_MAX);

    /* Close all open files except stdout and stderr. */
    for (fd = 0; fd < fd_count; fd++) {
        if (fd != STDOUT_FILENO && fd != STDERR_FILENO) {
            close(fd);
        }
    }

    /* Replace stdin with /dev/null. */
    if (open(_PATH_DEVNULL, O_RDONLY) < 0) {
        xlog(LOG_ERR, "error; unable to open " _PATH_DEVNULL ": %s\n",
             xstrerror(errno));
        exit(1);
    }

    /* Write our process id to the ancestor. */
    pid = getpid();
    fwrite(&pid, sizeof(pid), 1, stdout);
    fflush(stdout);
}

int
main(int argc, char *argv[])
{
    int err;
    int choice, i, count;
    int facility = LOG_DAEMON;
    struct fsm fsm;
    sigset_t sigset, oldset;
    struct timeval now, usdelay;
    struct timespec nsdelay, *timeout;
    fd_set readfds;
    struct sigaction action;
    int do_exit, input;
    int found, nfds;
    char *p;

    /* Determine the name of the executable. */
    if ((progname = strrchr(argv[0], '/')) == NULL) {
        progname = argv[0];
    } else {
        progname++;
    }

    /* Initialize the finite state machine. */
    fsm_init(&fsm);

    /* Parse the command-line options. */
    while ((choice = getopt(argc, argv, OPTIONS)) != EOF) {
        switch (choice) {
            case 'a':
                if ((err = parse_pvc(optarg,
                                     (struct sockaddr *)&fsm.atm_addr,
                                     sizeof(fsm.atm_addr))) != 0) {
                    xlog(LOG_ERR, "error: invalid ATM address: %s", optarg);
                    exit(1);
                }

                /* If the ATM interface number was specified then copy
                 * it to atm_itf. */
                if (fsm.atm_addr.sap_addr.itf >= 0) {
                    fsm.atm_itf = fsm.atm_addr.sap_addr.itf;
                }

                fsm.conntype = CT_PPPoE;
                break;

            case 'e':
                if (strcasecmp(optarg, "vcmux") == 0) {
                    fsm.encap = BR2684_ENCAPS_VC;
                } else if (strcasecmp(optarg, "llc") == 0) {
                    fsm.encap = BR2684_ENCAPS_LLC;
                } else {
                    xlog(LOG_ERR, "error: invalid encapsulation: %s", optarg);
                    exit(1);
                }

                break;

            case 'F':
                log_to_stderr = 1;
                fsm.daemonize = 0;
                break;

            case 'f':
                facility = -1;

                /* Look for a symbolic facility. */
                for (i = 0; facilitynames[i].c_name != NULL; i++) {
                    if (strcasecmp(facilitynames[i].c_name, optarg) == 0) {
                        facility = facilitynames[i].c_val;
                        break;
                    }
                }

                /* Complain if we didn't find the facility. */
                if (facility < 0) {
                    xlog(LOG_ERR, "error: unknown facility: %s", optarg);
                    exit(1);
                }

                break;

            case 'i':
                fsm.atm_itf = strtol(optarg, &p, 0);
                if (*p != 0 || fsm.atm_itf < 0) {
                    xlog(LOG_ERR, "error: invalid ATM interface number: %s",
                         optarg);
                    exit(1);
                }
                break;

            case 'k':
		fsm.exit_on_error = 0;
                break;

            case 'm':
                fsm.mac_addr = optarg;
                break;

            case 'p':
                pid_file = optarg;
                break;

            case 's':
                fsm.detach_on_signal = 1;
                break;

            case 'u':
                fsm.mod_list = optarg;
                break;

            case 'v':
                /* Look for a symbolic priority. */
                found = 0;
                for (i = 0; prioritynames[i].c_name != NULL; i++) {
                    if (strcasecmp(prioritynames[i].c_name, optarg) == 0) {
                        verbosity = prioritynames[i].c_val;
                        found = 1;
                        break;
                    }
                }

                if (!found) {
                    /* Try a numeric priority. */
                    verbosity = strtol(optarg, &p, 0);
                    if (*p == 0 && verbosity >= LOG_EMERG &&
                        verbosity <= LOG_DEBUG) {
                        found = 1;
                    }
                }

                /* Complain if invalid priority given. */
                if (!found) {
                    verbosity = LOG_ERR;
                    xlog(LOG_ERR, "error: invalid verbosity level: %s",
                         optarg);
                    exit(1);
                }

                break;

            case 'h':
                usage();
                exit(0);

            case 'V':
                fprintf(stderr, "%s (" PACKAGE ") version " VERSION "\n",
                        progname);
                exit(0);

            case '?':
                usage();
                exit(1);

            default:
                fprintf(stderr, "%s: error: unhandled option -%c\n",
                        progname, choice);
                exit(1);
        }
    }

    /* Record pppd's argc and argv. */
    fsm.pppd_argc = argc - optind;
    fsm.pppd_argv = argv + optind;

    /* Daemonize, sort of.  The descendant will do all of the
     * interesting work, but the ancestor process will hang around
     * until the ppp connection is established or fails. */
    if (fsm.daemonize) {
        xlog(LOG_DEBUG, "daemonizing");
        daemonize();
    }

    /* Set up syslog. */
    openlog(progname, 0, facility);

    /* Determine the path to modprobe. */
    if (fsm.mod_list) {
        if ((err = find_modprobe()) != 0) {
            xlog(LOG_ERR, "error: unable to locate modprobe: %s",
                 xstrerror(err));
            exit(1);
        }

        xlog(LOG_DEBUG, "modprobe is %s", path_modprobe);
    }

    /* Block various signals and set their signal handlers. */
    ASSERT(sigemptyset(&sigset) == 0);
    for (i = 0; blocked_signals[i]; i++) {
        ASSERT(sigaddset(&sigset, blocked_signals[i]) == 0);
        ASSERT(signal(blocked_signals[i], handle_signal) != SIG_ERR);
    }

    /* Use sigaction to register SIGCHLD so that we can indicate that
     * we don't care about STOP signals. */
    ASSERT(sigaddset(&sigset, SIGCHLD) == 0);
    memset(&action, 0, sizeof(struct sigaction));
    action.sa_handler = handle_signal;
    action.sa_flags = SA_NOCLDSTOP;
    sigaction(SIGCHLD, &action , NULL);

    ASSERT(sigprocmask(SIG_BLOCK, &sigset, &oldset) == 0);

    /* Do the state machine thing. */
    do_exit = 0;
    WAKESETNOW(&fsm.wake);
    for (;;) {
	/* Determine how long to wait for something to happen. */
        if (fsm.wake.tv_sec < 0 && fsm.wake.tv_usec < 0) {
            /* Wait indefinitely. */
            timeout = NULL;
        } else if (fsm.wake.tv_sec == 0 && fsm.wake.tv_usec == 0) {
            /* Don't wait at all. */
            memset(&nsdelay, 0, sizeof(nsdelay));
            timeout = &nsdelay;
        } else {
            /* Get the current time. */
            xgettimeofday(&now);
            if (timercmp(&fsm.wake, &now, <)) {
                /* Wake time has passed; don't wait at all. */
                memset(&nsdelay, 0, sizeof(nsdelay));
                timeout = &nsdelay;
            } else {
                /* Wait until wake time comes. */
                timersub(&fsm.wake, &now, &usdelay);
                nsdelay.tv_sec = usdelay.tv_sec;
                nsdelay.tv_nsec = usdelay.tv_usec * 1000;
                timeout = &nsdelay;
                xlog(LOG_DEBUG, "pausing for %d.%06ds",
                     usdelay.tv_sec, usdelay.tv_usec);
            }
        }

	/* Set up the one file descriptor set we use. */
	FD_ZERO(&readfds);
	if (fsm.pcap_sock < 0) {
	    nfds = 0;
	} else {
	    FD_SET(fsm.pcap_sock, &readfds);
	    nfds = fsm.pcap_sock + 1;
	}

        /* Wait for something to happen. */
	ASSERT(last_signum == 0);
        count = pselect(nfds, &readfds, NULL, NULL, timeout, &oldset);

	switch (count) {
            case -1:
                /* Die for any error other than an interrupt. */
                if (errno != EINTR) {
                    xlog(LOG_ERR, "pselect failed: %s", xstrerror(errno));
                    exit(1);
                }

                /* A signal handler should have been called. */
                ASSERT(last_signum != 0);
                break;

            case 0:
                /* No signal handler should have been called. */
                ASSERT(last_signum == 0);
                ASSERT(timeout);
                xgettimeofday(&now);

                /* If the timer has expired early then go back to sleep. */
                if (timercmp(&now, &fsm.wake, <)) {
                    timersub(&fsm.wake, &now, &usdelay);
                    if (usdelay.tv_sec != 0 || usdelay.tv_usec > 1000) {
                        continue;
                    }
                }

                /* Otherwise translate the timeout into a SIGALRM. */
                last_signum = SIGALRM;
                break;

            case 1:
                /* No signal handler should have been called. */
                ASSERT(last_signum == 0);

                /* We shouldn't be snooping if we have a session id. */
                ASSERT(fsm.bridge.session_id == 0);

                /* Receive a packet. */
                err = pcap_recv(fsm.pcap_sock, &fsm.bridge);
                if (err) {
                    xlog(LOG_WARNING, "unable to read from packet capture socket: %s",
                         xstrerror(err));
                    close(fsm.pcap_sock);
                    fsm.pcap_sock = -1;
                } else if (fsm.bridge.session_id) {
                    xlog(LOG_INFO, "PPPoE session id 0x%hx",
                         fsm.bridge.session_id);

                    /* We now know the PPPoE session id, so we no
                     * longer need to snoop on packets. */
                    if (close(fsm.pcap_sock) < 0) {
                        xlog(LOG_WARNING, "unable to close packet capture socket: %s",
                             xstrerror(errno));
                    }
                    fsm.pcap_sock = -1;

                    /* Co-opt a signal that we can't handle. */
                    last_signum = SIGCONT;
                    break;
                }

                /* We haven't learned anything.  Loop again. */
                continue;
        }

        /* Kill pppd in response to SIGHUP, SIGINT or SIGTERM. */
        if (last_signum == SIGHUP || last_signum == SIGINT ||
            last_signum == SIGTERM) {
            /* We've handled the signal. */
            last_signum = 0;

            /* Shut down on SIGINT or SIGTERM. */
            if (last_signum != SIGHUP) {
                xlog(LOG_INFO, "exiting on %s",
                     last_signum == SIGINT ? "SIGINT" : "SIGTERM");
            }

            /* If pppd is running then kill it. */
            if (fsm.pppd_pid != (pid_t)-1) {
                if (kill(fsm.pppd_pid, SIGTERM) < 0) {
                    xlog(LOG_ERR, "error: unable to kill child process: %s",
                         xstrerror(errno));
                    fsm.pppd_pid = (pid_t)-1;
                }
            }

            /* If this is the second time we've been asked to quit
             * then don't wait for pppd. */
            if (last_signum != SIGHUP) {
                if (do_exit++ == 0) {
                    fsm.exit_code = 0;
                } else {
                    fsm.pppd_pid = (pid_t)-1;
                    fsm.exit_code = 1;
                }
            }

            /* Wait for pppd to die if we're going to. */
            if (fsm.pppd_pid != (pid_t)-1) {
                SET_STATE(fsm.state, FSM_REAP);
                WAKESETNONE(&fsm.wake);
                continue;
            }

            /* Otherwise just clean up as best we can and quit. */
            fsm_cleanup(&fsm, CU_MODULES | CU_BRIDGE);
            exit(0);
        }

        /* Detach on SIGUSR1. */
        if (last_signum == SIGUSR1) {
            if (fsm.daemonize && fsm.detach_on_signal) {
		fsm_detach(&fsm);
            }
	    last_signum = 0;
	    continue;
        }

	/* Consume the signal. */
	input = last_signum;
	last_signum = 0;

        switch (fsm.state) {
            case FSM_START:
                fsm_start(&fsm, input);
                break;

            case FSM_FIND_STATUS:
                fsm_find_status(&fsm, input);
                break;

            case FSM_CARRIER:
                fsm_carrier(&fsm, input);
                break;

            case FSM_PPPD:
		fsm_pppd(&fsm, input);
                break;

            case FSM_SESSION:
		fsm_session(&fsm, input);
                break;

            case FSM_UP:
		fsm_up(&fsm, input);
                break;

            case FSM_PADT:
		fsm_padt(&fsm, input);
                break;

            case FSM_REAP:
		fsm_reap(&fsm, input);
                break;

            default:
                xlog(LOG_ERR, "error: unknown state: %d", fsm.state);
                exit(1);
        }
    }
}
